
/** @file
    @brief		sNX
    @author		t`
    @since		2003.07.01
*/


//==========================================================================
// INCLUDE
//==========================================================================
#include "LunaBase.h"

#include "LunaMatrix.h"
#include "LunaMath.h"

namespace lunadxmath{
//=============================================================================
/**
	RXgN^
*/
//=============================================================================
CMatrix::CMatrix()
{
	Init();
}

//=============================================================================
/**
	RXgN^

	@param Mat	[in] ps
*/
//=============================================================================
CMatrix::CMatrix( CMatrix &Mat )
{
	Init();
	Set( Mat.Get() );
}

//=============================================================================
/**
	s̏
*/
//=============================================================================
void CMatrix::Init( void )
{
	Uint32 Addr = (Uint32)Data;
	Addr += 0x0000000F;
	Addr &= 0xFFFFFFF0;
	pM = (D3DXMATRIX*)Addr;
}

//=============================================================================
/**
	s̃AhX擾

	@return		s̃AhX
*/
//=============================================================================
D3DXMATRIX *CMatrix::Get( void )
{
	return pM;
}

//=============================================================================
/**
	s̃Rs[

	@param pOut [out] Rs[s
*/
//=============================================================================
void CMatrix::Get( D3DXMATRIX *pOut )
{
	MemoryCopy( pOut, pM, sizeof(D3DXMATRIX) );
}

//=============================================================================
/**
	s̃Rs[

	@param pIn	[in] Rs[s
*/
//=============================================================================
void CMatrix::Set( D3DXMATRIX *pIn )
{
	MemoryCopy( pM, pIn, sizeof(D3DXMATRIX) );
}

//=============================================================================
/**
	s̒Pʉ
*/
//=============================================================================
void CMatrix::Identity( void )
{
	pM->m[0][0] = 1.0f;
	pM->m[0][1] = 0.0f;
	pM->m[0][2] = 0.0f;
	pM->m[0][3] = 0.0f;

	pM->m[1][0] = 0.0f;
	pM->m[1][1] = 1.0f;
	pM->m[1][2] = 0.0f;
	pM->m[1][3] = 0.0f;

	pM->m[2][0] = 0.0f;
	pM->m[2][1] = 0.0f;
	pM->m[2][2] = 1.0f;
	pM->m[2][3] = 0.0f;

	pM->m[3][0] = 0.0f;
	pM->m[3][1] = 0.0f;
	pM->m[3][2] = 0.0f;
	pM->m[3][3] = 1.0f;
}

//=============================================================================
/**
	s̓]u

	@param pIn	[in] s
*/
//=============================================================================
void CMatrix::Transpose( CMatrix *pIn )
{
	::D3DXMatrixTranspose( pM, pIn->pM );
}

//=============================================================================
/**
	s̋ts

	@param pIn	[in] s
*/
//=============================================================================
void CMatrix::Inverse( CMatrix *pIn )
{
	::D3DXMatrixInverse( pM, NULL, pIn->pM );
}

//=============================================================================
/**
	s̉]vf̐K

	@param pIn	[in] s
*/
//=============================================================================
void CMatrix::Normalize( CMatrix *pIn )
{
	CVector3 v1( pIn->pM->m[0][0], pIn->pM->m[1][0], pIn->pM->m[2][0] );
	CVector3 v2( pIn->pM->m[0][1], pIn->pM->m[1][1], pIn->pM->m[2][1] );
	CVector3 v3( pIn->pM->m[0][2], pIn->pM->m[1][2], pIn->pM->m[2][2] );

	v1.Normalize();
	v2.Normalize();
	v3.Normalize();

	pM->m[0][0] = v1.x;
	pM->m[1][0] = v1.y;
	pM->m[2][0] = v1.z;
	pM->m[0][1] = v2.x;
	pM->m[1][1] = v2.y;
	pM->m[2][1] = v2.z;
	pM->m[0][2] = v3.x;
	pM->m[1][2] = v3.y;
	pM->m[2][2] = v3.z;
}

//=============================================================================
/**
	s̍

	@param pIn1		[in] sP
	@param pIn2		[in] sQ
*/
//=============================================================================
void CMatrix::Multiply( CMatrix *pIn1, CMatrix *pIn2 )
{
	::D3DXMatrixMultiply( pM, pIn1->pM, pIn2->pM );
}

//=============================================================================
/**
	s̓]us

*/
//=============================================================================
void CMatrix::Transpose( void )
{
	::D3DXMatrixTranspose( pM, pM );
}

//=============================================================================
/**
	s̋ts

*/
//=============================================================================
void CMatrix::Inverse( void )
{
	::D3DXMatrixInverse( pM, NULL, pM );
}

//=============================================================================
/**
	s̉]vf̐K

*/
//=============================================================================
void CMatrix::Normalize( void )
{
	Normalize( this );
}

//=============================================================================
/**
	s̍

	@param pIn	[in] s
*/
//=============================================================================
void CMatrix::Multiply( CMatrix *pIn )
{
	::D3DXMatrixMultiply( pM, pM, pIn->pM );
}

//=============================================================================
/**
	w]px

	@param Angle	[in] ]px
*/
//=============================================================================
void CMatrix::RotationX( Sint32 Angle )
{
	Float c = LunaMath::Cos( Angle );
	Float s = LunaMath::Sin( Angle );

	pM->m[0][0] = 1;  pM->m[0][1] =  0;  pM->m[0][2] = 0;  pM->m[0][3] = 0;
	pM->m[1][0] = 0;  pM->m[1][1] =  c;  pM->m[1][2] = s;  pM->m[1][3] = 0;
	pM->m[2][0] = 0;  pM->m[2][1] = -s;  pM->m[2][2] = c;  pM->m[2][3] = 0;
	pM->m[3][0] = 0;  pM->m[3][1] =  0;  pM->m[3][2] = 0;  pM->m[3][3] = 1;
}

//=============================================================================
/**
	x]px

	@param Angle	[in] ]px
*/
//=============================================================================
void CMatrix::RotationY( Sint32 Angle )
{
	Float c = LunaMath::Cos( Angle );
	Float s = LunaMath::Sin( Angle );

	pM->m[0][0] = c;  pM->m[0][1] = 0;  pM->m[0][2] = -s;  pM->m[0][3] = 0;
	pM->m[1][0] = 0;  pM->m[1][1] = 1;  pM->m[1][2] =  0;  pM->m[1][3] = 0;
	pM->m[2][0] = s;  pM->m[2][1] = 0;  pM->m[2][2] =  c;  pM->m[2][3] = 0;
	pM->m[3][0] = 0;  pM->m[3][1] = 0;  pM->m[3][2] =  0;  pM->m[3][3] = 1;
}

//=============================================================================
/**
	y]px

	@param Angle	[in] ]px
*/
//=============================================================================
void CMatrix::RotationZ( Sint32 Angle )
{
	Float c = LunaMath::Cos( Angle );
	Float s = LunaMath::Sin( Angle );

	pM->m[0][0] =  c;  pM->m[0][1] = s;  pM->m[0][2] = 0;  pM->m[0][3] = 0;
	pM->m[1][0] = -s;  pM->m[1][1] = c;  pM->m[1][2] = 0;  pM->m[1][3] = 0;
	pM->m[2][0] =  0;  pM->m[2][1] = 0;  pM->m[2][2] = 1;  pM->m[2][3] = 0;
	pM->m[3][0] =  0;  pM->m[3][1] = 0;  pM->m[3][2] = 0;  pM->m[3][3] = 1;
}

//=============================================================================
/**
	ywx]s𐶐

	@param AngleX	[in] w]px
	@param AngleY	[in] x]px
	@param AngleZ	[in] y]px
*/
//=============================================================================
void CMatrix::RotationZXY( Sint32 AngleX, Sint32 AngleY, Sint32 AngleZ )
{
	AngleX >>= 1;
	AngleY >>= 1;
	AngleZ >>= 1;

	Float SinAx = LunaMath::Sin( AngleX );
	Float SinAy = LunaMath::Sin( AngleY );
	Float SinAz = LunaMath::Sin( AngleZ );
	Float CosAx = LunaMath::Cos( AngleX );
	Float CosAy = LunaMath::Cos( AngleY );
	Float CosAz = LunaMath::Cos( AngleZ ); 

	Float x = (SinAx * CosAy * CosAz) - (CosAx * SinAy * SinAz);
	Float y = (CosAx * SinAy * CosAz) + (SinAx * CosAy * SinAz);
	Float z = (CosAx * CosAy * SinAz) - (SinAx * SinAy * CosAz);
	Float w = (CosAx * CosAy * CosAz) + (SinAx * SinAy * SinAz);

	Float xx = x * x;
	Float yy = y * y;
	Float zz = z * z;
	Float xy = x * y;
	Float xz = x * z;
	Float yz = y * z;
	Float wx = w * x;
	Float wy = w * y;
	Float wz = w * z;

	pM->m[0][0] = 1 - (2 * (yy + zz));
	pM->m[0][1] =     (2 * (xy - wz));
	pM->m[0][2] =     (2 * (xz + wy));
	pM->m[0][3] = 0;

	pM->m[1][0] =     (2 * (xy + wz));
	pM->m[1][1] = 1 - (2 * (xx + zz));
	pM->m[1][2] =     (2 * (yz - wx));
	pM->m[1][3] = 0;

	pM->m[2][0] =     (2 * (xz - wy));
	pM->m[2][1] =     (2 * (yz + wx));
	pM->m[2][2] = 1 - (2 * (xx + yy));
	pM->m[2][3] = 0;

	pM->m[3][0] = 0;
	pM->m[3][1] = 0;
	pM->m[3][2] = 0;
	pM->m[3][3] = 1;
}

//=============================================================================
/**
	Cӎ]

	@param Angle	[in] px
	@param pAxis	[in] ]
*/
//=============================================================================
void CMatrix::RotationAxis( Sint32 Angle, class CVector3 *pAxis )
{
	CVector3 Axis;
	Axis.Normalize( pAxis );

	Float c = LunaMath::Cos( Angle );
	Float s = LunaMath::Sin( Angle );

	pM->m[0][0] = (1.0f-c)*(Axis.x*Axis.x) + c;
	pM->m[0][1] = (1.0f-c)*(Axis.x*Axis.y) - (Axis.z*s);
	pM->m[0][2] = (1.0f-c)*(Axis.z*Axis.x) + (Axis.y*s);
	pM->m[0][3] = 0.0f;

	pM->m[1][0] = (1.0f-c)*(Axis.x*Axis.y) + (Axis.z*s);
	pM->m[1][1] = (1.0f-c)*(Axis.y*Axis.y) + c;
	pM->m[1][2] = (1.0f-c)*(Axis.y*Axis.z) - (Axis.x*s);
	pM->m[1][3] = 0.0f;

	pM->m[2][0] = (1.0f-c)*(Axis.z*Axis.x) - (Axis.y*s);
	pM->m[2][1] = (1.0f-c)*(Axis.y*Axis.z) + (Axis.x*s);
	pM->m[2][2] = (1.0f-c)*(Axis.z*Axis.z) + c;
	pM->m[2][3] = 0.0f;

	pM->m[3][0] = 0.0f;
	pM->m[3][1] = 0.0f;
	pM->m[3][2] = 0.0f;
	pM->m[3][3] = 1.0f;
}

//=============================================================================
/**
	R]s̐

	@param rotx	[in] w]px
	@param roty	[in] x]px
	@param rotz	[in] y]px
*/
//=============================================================================
void CMatrix::Rotation( Sint32 rotx, Sint32 roty, Sint32 rotz )
{
	Float xx, yy, zz, xy, xz, yz, wx, wy, wz;
	Float sx, sy, sz, cx, cy, cz;
	Float x, y, z, w;

	LunaMath::SinCos( rotx/2, &sx, &cx );
	LunaMath::SinCos( roty/2, &sy, &cy );
	LunaMath::SinCos( rotz/2, &sz, &cz );

	x = (sx * cy * cz) - (cx * sy * sz);
	y = (cx * sy * cz) + (sx * cy * sz);
	z = (cx * cy * sz) - (sx * sy * cz);
	w = (cx * cy * cz) + (sx * sy * sz);

	xx = x * x;
	yy = y * y;
	zz = z * z;
	xy = x * y;
	xz = x * z;
	yz = y * z;
	wx = w * x;
	wy = w * y;
	wz = w * z;

	pM->m[0][0] = 1 - (2 * (yy+zz));
	pM->m[0][1] = 2 * (xy - wz);
	pM->m[0][2] = 2 * (xz + wy);
	pM->m[0][3] = 0;

	pM->m[1][0] = 2 * (xy + wz);
	pM->m[1][1] = 1 - (2 * (xx+zz));
	pM->m[1][2] = 2 * (yz-wx);
	pM->m[1][3] = 0;

	pM->m[2][0] = 2 * (xz-wy);
	pM->m[2][1] = 2 * (yz+wx);
	pM->m[2][2] = 1 - (2 * (xx+yy));
	pM->m[2][3] = 0;

	pM->m[3][0] = 0;
	pM->m[3][1] = 0;
	pM->m[3][2] = 0;
	pM->m[3][3] = 1;
}

//=============================================================================
/**
	xNgɉ]

	@param pIn1		[in] n_xNg
	@param pIn2		[in] I_xNg
	@param pAxis	[in] ]
*/
//=============================================================================
void CMatrix::RotateVector( class CVector3 *pIn1, class CVector3 *pIn2, class CVector3 *pAxis )
{
	// ]߂
	CVector3 vCross;
	if ( pAxis == NULL )
	{
		vCross.CrossProduct( pIn1, pIn2 );
	}
	else
	{
		vCross = *pAxis;
	}

	// ]p߂
	Float cs = pIn1->DotProduct( pIn2 ) / (pIn1->Length() * pIn2->Length());

	// s񐶐
	Float sn = (Float)sqrt( 1 - (cs * cs) );

	pM->m[0][0] = (vCross.x*vCross.x) * (1-cs)+ cs;
	pM->m[1][0] = (vCross.y*vCross.x) * (1-cs)- (vCross.z*sn);
	pM->m[2][0] = (vCross.z*vCross.x) * (1-cs)+ (vCross.y*sn);
	pM->m[3][0] = 0;

	pM->m[0][1] = (vCross.x*vCross.y) * (1-cs)+ (vCross.z*sn);
	pM->m[1][1] = (vCross.y*vCross.y) * (1-cs)+ cs;
	pM->m[2][1] = (vCross.z*vCross.y) * (1-cs)- (vCross.x*sn);
	pM->m[3][1] = 0;

	pM->m[0][2] = (vCross.x*vCross.z) * (1-cs)- (vCross.y*sn);
	pM->m[1][2] = (vCross.y*vCross.z) * (1-cs)+ (vCross.x*sn);
	pM->m[2][2] = (vCross.z*vCross.z) * (1-cs)+ cs;
	pM->m[3][2] = 0;

	pM->m[0][3] = 0;
	pM->m[1][3] = 0;
	pM->m[2][3] = 0;
	pM->m[3][3] = 1;
}

//=============================================================================
/**
	gks񐶐

	@param sx	[in] wg嗦
	@param sy	[in] xg嗦
	@param sz	[in] yg嗦
*/
//=============================================================================
void CMatrix::Scaling( Float sx, Float sy, Float sz )
{
	pM->m[0][0] = sx;  pM->m[0][1] =  0;  pM->m[0][2] =  0;  pM->m[0][3] = 0;
	pM->m[1][0] =  0;  pM->m[1][1] = sy;  pM->m[1][2] =  0;  pM->m[1][3] = 0;
	pM->m[2][0] =  0;  pM->m[2][1] =  0;  pM->m[2][2] = sz;  pM->m[2][3] = 0;
	pM->m[3][0] =  0;  pM->m[3][1] =  0;  pM->m[3][2] =  0;  pM->m[3][3] = 1;
}

//=============================================================================
/**
	ړs񐶐

	@param px	[in] wʒu
	@param py	[in] xʒu
	@param pz	[in] yʒu
*/
//=============================================================================
void CMatrix::Translation( Float px, Float py, Float pz )
{
	pM->m[0][0] = 1;  pM->m[0][1] = 0;  pM->m[0][2] = 0;  pM->m[0][3] = 0;
	pM->m[1][0] = 0;  pM->m[1][1] = 1;  pM->m[1][2] = 0;  pM->m[1][3] = 0;
	pM->m[2][0] = 0;  pM->m[2][1] = 0;  pM->m[2][2] = 1;  pM->m[2][3] = 0;

	pM->m[3][0] = px;
	pM->m[3][1] = py;
	pM->m[3][2] = pz;
	pM->m[3][3] = 1;
}

//=============================================================================
/**
	ϊps񐶐

	@param fov		[in] p
	@param near_z	[in] ߐڃNbv
	@param far_z	[in] Nbv
	@param aspect	[in] AXyNg
*/
//=============================================================================
void CMatrix::Perspective( Float fov, Float near_z, Float far_z, Float aspect )
{
	::D3DXMatrixPerspectiveFovLH( pM, fov, aspect, near_z, far_z );
/*
	Float fTemp = fov * 0.5f;
	fTemp = F(cos(fTemp)) / F(sin(fTemp));

	pM->m[0][0] = fTemp * aspect;
	pM->m[0][1] = 0.0f;
	pM->m[0][2] = 0.0f;
	pM->m[0][3] = 0.0f;

	pM->m[1][0] = 0.0f;
	pM->m[1][1] = fTemp;
	pM->m[1][2] = 0.0f;
	pM->m[1][3] = 0.0f;

	pM->m[2][0] = 0.0f;
	pM->m[2][1] = 0.0f;
	pM->m[2][2] = far_z / (far_z - near_z);
	pM->m[2][3] = 1.0f;

	pM->m[3][0] = 0.0f;
	pM->m[3][1] = 0.0f;
	pM->m[3][2] = -near_z * fTemp;
	pM->m[3][3] = 0.0f;
*/
}

//=============================================================================
/**
	seps񐶐

	@param w		[in] 
	@param h		[in] c
	@param near_z	[in] ߐڃNbv
	@param far_z	[in] Nbv
*/
//=============================================================================
void CMatrix::Ortho( Float w, Float h, Float near_z, Float far_z )
{
	::D3DXMatrixOrthoLH( pM, w, h, near_z, far_z );
/*
	pM->m[0][0] = 2.0f / w;
	pM->m[0][1] = 0.0f;
	pM->m[0][2] = 0.0f;
	pM->m[0][3] = 0.0f;

	pM->m[1][0] = 0.0f;
	pM->m[1][1] = 2.0f / h;
	pM->m[1][2] = 0.0f;
	pM->m[1][3] = 0.0f;

	pM->m[2][0] = 0.0f;
	pM->m[2][1] = 0.0f;
	pM->m[2][2] = 1.0f / (far_z - near_z);
	pM->m[2][3] = 0.0f;

	pM->m[3][0] = 0.0f;
	pM->m[3][1] = 0.0f;
	pM->m[3][2] = near_z / (near_z - far_z);
	pM->m[3][3] = 1.0f;
*/
}

//=============================================================================
/**
	Cӓ_s쐬

	@param pEye	[in] _
	@param pAt	[in] _
	@param pUp	[in] 
*/
//=============================================================================
void CMatrix::LookAt( class CVector3 *pEye, class CVector3 *pAt, class CVector3 *pUp )
{
	::D3DXMatrixLookAtLH( pM, pEye, pAt, pUp );
}

//=============================================================================
/**
	Iy[^[
*/
//=============================================================================
CMatrix &CMatrix::operator = ( const CMatrix &In )
{
	MemoryCopy( pM, In.pM, sizeof(D3DXMATRIX) );

	return *this;
}

//=============================================================================
/**
	Iy[^[
*/
//=============================================================================
CMatrix &CMatrix::operator = ( const class CQuaternion &In )
{
	Float xx = In.x * In.x;
	Float yy = In.y * In.y;
	Float zz = In.z * In.z;
	Float xy = In.x * In.y;
	Float xz = In.x * In.z;
	Float yz = In.y * In.z;
	Float wx = In.w * In.x;
	Float wy = In.w * In.y;
	Float wz = In.w * In.z;

	pM->m[0][0] = 1.0f - 2.0f * (yy + zz);
	pM->m[0][1] =        2.0f * (xy - wz);
	pM->m[0][2] =        2.0f * (xz + wy);
	pM->m[0][3] = 0.0f;

	pM->m[1][0] =        2.0f * (xy + wz);
	pM->m[1][1] = 1.0f - 2.0f * (xx + zz);
	pM->m[1][2] =        2.0f * (yz - wx);
	pM->m[1][3] = 0.0f;

	pM->m[2][0] =        2.0f * (xz - wy);
	pM->m[2][1] =        2.0f * (yz + wx);
	pM->m[2][2] = 1.0f - 2.0f * (xx + yy);
	pM->m[2][3] = 0.0f;

	pM->m[3][0] = 0.0f;
	pM->m[3][1] = 0.0f;
	pM->m[3][2] = 0.0f;
	pM->m[3][3] = 1.0f;

	return *this;
}
}
